import collections.abc as abc
from functools import partial
from torch.utils.data._utils.collate import default_collate
from torch.utils.data import DataLoader
from typing import Union, Callable, Tuple
import copy

import numpy as np
import torch
from opacus import PrivacyEngine
from opacus.validators import ModuleValidator


def opacus_model_fix(model: torch.nn.Module, params: dict) -> torch.nn.Module:
    """
    Function to detect components of the model that are not compatible with Opacus differentially private training, and replacing with compatible components
    or raising an exception when a fix cannot be handled by Opacus.

    Args:
        model (torch.nn.Module): The model to be trained.
        params (dict): Training parameters.

    Returns:
        torch.nn.Module: Model, with potentially some components replaced with ones compatible with Opacus.
    """
    # use opacus to detect issues with model
    opacus_errors_detected = ModuleValidator.validate(model, strict=False)

    if not params["differential_privacy"]["allow_opacus_model_fix"]:
        assert (
            opacus_errors_detected == []
        ), f"Training parameters are set to not allow Opacus to try to fix incompatible model components, and the following issues were detected: {opacus_errors_detected}"
    elif opacus_errors_detected != []:
        print(
            f"Allowing Opacus to try and patch the model due to the following issues: {opacus_errors_detected}"
        )
        print()
        model = ModuleValidator.fix(model)
        # If the fix did not work, raise an exception
        ModuleValidator.validate(model, strict=True)
    return model


def prep_for_opacus_training(
    model: torch.nn.Module,
    optimizer: torch.optim.Optimizer,
    train_dataloader: DataLoader,
    params: dict,
) -> Tuple[torch.nn.Module, torch.optim.Optimizer, DataLoader, PrivacyEngine]:
    """
    Function to prepare the model, optimizer, and dataloader for differentially private training using Opacus.

    Args:
        model (torch.nn.Module): The model to be trained.
        optimizer (torch.optim.Optimizer): The optimizer to be used for training.
        train_dataloader (DataLoader): The dataloader for the training data.
        params (dict): Training parameters.

    Returns:
        Tuple[torch.nn.Module, torch.optim.Optimizer, DataLoader, PrivacyEngine]: Model, optimizer, dataloader, and privacy engine.
    """

    privacy_engine = PrivacyEngine(
        accountant=params["differential_privacy"]["accountant"],
        secure_mode=params["differential_privacy"]["secure_mode"],
    )

    if not "epsilon" in params["differential_privacy"]:
        model, optimizer, train_dataloader = privacy_engine.make_private(
            module=model,
            optimizer=optimizer,
            data_loader=train_dataloader,
            noise_multiplier=params["differential_privacy"]["noise_multiplier"],
            max_grad_norm=params["differential_privacy"]["max_grad_norm"],
        )
    else:
        (model, optimizer, train_dataloader) = privacy_engine.make_private_with_epsilon(
            module=model,
            optimizer=optimizer,
            data_loader=train_dataloader,
            max_grad_norm=params["differential_privacy"]["max_grad_norm"],
            epochs=params["num_epochs"],
            target_epsilon=params["differential_privacy"]["epsilon"],
            target_delta=params["differential_privacy"]["delta"],
        )
    return model, optimizer, train_dataloader, privacy_engine


def build_empty_batch_value(
    sample: Union[torch.Tensor, np.ndarray, abc.Mapping, abc.Sequence, int, float, str]
):
    """
    Build an empty batch value from a sample. This function is used to create a placeholder for empty batches in an iteration. Inspired from https://github.com/pytorch/pytorch/blob/main/torch/utils/data/_utils/collate.py#L108. The key difference is that pytorch `collate` has to traverse batch of objects AND unite its fields to lists, while this function traverse a single item AND creates an "empty" version of the batch.

    Args:
        sample (Union[torch.Tensor, np.ndarray, abc.Mapping, abc.Sequence, int, float, str]): A sample from the dataset.

    Raises:
        TypeError: If the data type is not supported.

    Returns:
        Union[torch.Tensor, np.ndarray, abc.Mapping, abc.Sequence, int, float, str]: An empty batch value.
    """
    if isinstance(sample, torch.Tensor):
        # Create an empty tensor with the same shape except for the zeroed batch dimension.
        return torch.empty((0,) + sample.shape)
    elif isinstance(sample, np.ndarray):
        # Create an empty tensor from a numpy array, also with the zeroed batch dimension.
        return torch.empty((0,) + sample.shape, dtype=torch.from_numpy(sample).dtype)
    elif isinstance(sample, abc.Mapping):
        # Recursively handle dictionary-like objects.
        return {key: build_empty_batch_value(value) for key, value in sample.items()}
    elif isinstance(sample, tuple) and hasattr(sample, "_fields"):  # namedtuple
        return type(sample)(*(build_empty_batch_value(item) for item in sample))
    elif isinstance(sample, abc.Sequence) and not isinstance(sample, str):
        # Handle lists and tuples, but exclude strings.
        return [build_empty_batch_value(item) for item in sample]
    elif isinstance(sample, (int, float, str)):
        # Return an empty list for basic data types.
        return []
    else:
        raise TypeError(f"Unsupported data type: {type(sample)}")


def empty_collate(
    item_example: Union[
        torch.Tensor, np.ndarray, abc.Mapping, abc.Sequence, int, float, str
    ]
) -> Callable:
    """
    Creates a new collate function that behave same as default pytorch one,
    but can process the empty batches.

    Args:
        item_example (Union[torch.Tensor, np.ndarray, abc.Mapping, abc.Sequence, int, float, str]): An example item from the dataset.

    Returns:
        Callable: function that should replace dataloader collate: `dataloader.collate_fn = empty_collate(...)`
    """

    def custom_collate(batch, _empty_batch_value):
        if len(batch) > 0:
            return default_collate(batch)  # default behavior
        else:
            return copy.copy(_empty_batch_value)

    empty_batch_value = build_empty_batch_value(item_example)

    return partial(custom_collate, _empty_batch_value=empty_batch_value)



from typing import Optional, Union, Dict
import os, datetime, subprocess, sys
from copy import deepcopy
import random
from packaging.version import Version

import numpy as np
import torch
import SimpleITK as sitk
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def checkPatchDivisibility(patch_size: np.ndarray, number: Optional[int] = 16) -> bool:
    """
    This function checks the divisibility of a numpy array or integer for architectural integrity

    Args:
        patch_size (np.ndarray): The patch size for checking.
        number (Optional[int], optional): The number to check divisibility for. Defaults to 16.

    Returns:
        bool: If all elements of array are divisible or not, after taking 2D patches into account.
    """
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]
    # for 2D, don't check divisibility of first dimension
    elif patch_size_to_check[0] == 1:
        patch_size_to_check = patch_size_to_check[1:]
    if np.count_nonzero(np.remainder(patch_size_to_check, number)) > 0:
        return False

    # adding check to address https://github.com/mlcommons/GaNDLF/issues/53
    # there is quite possibly a better way to do this
    unique = np.unique(patch_size_to_check)
    if (unique.shape[0] == 1) and (unique[0] < number):
        return False
    return True


def get_date_time() -> str:
    """
    Get a well-parsed date string

    Returns:
        str: The date in format YYYY/MM/DD::HH:MM:SS
    """
    now = datetime.datetime.now().strftime("%Y/%m/%d::%H:%M:%S")
    return now


def get_unique_timestamp() -> str:
    """
    Get a well-parsed timestamp string to be used for unique filenames

    Returns:
        str: The date in format YYYYMMDD_HHMMSS
    """
    now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    return now


def get_filename_extension_sanitized(filename: str) -> str:
    """
    This function returns the extension of the filename with leading and trailing characters removed.
    Args:
        filename (str): The filename to be processed.
    Returns:
        str: The filename with extension removed.
    """
    _, ext = os.path.splitext(filename)
    # if .gz or .nii file is detected, always return .nii.gz
    if (ext == ".gz") or (ext == ".nii"):
        ext = ".nii.gz"
    return ext


def version_check(version_from_config: Dict[str, str], version_to_check: str) -> bool:
    """
    This function checks if the version of the config file is compatible with the version of the code.

    Args:
        version_from_config (Dict[str, str]): The version from the config file which contains minimum and maximum versions.
        version_to_check (str): The version of the code or model to check.

    Returns:
        bool: If the version of the config file is compatible with the version of the code.
    """
    versions_from_config = {}
    for key_to_check in ["minimum", "maximum"]:
        versions_from_config[key_to_check] = version_from_config.get(key_to_check, None)
        assert (
            versions_from_config[key_to_check] is not None
        ), f"Version {key_to_check} is not specified in the config file"

    # create Version objects
    version_to_check_obj = Version(version_to_check)
    min_ver_obj = Version(versions_from_config["minimum"])
    max_ver_obj = Version(versions_from_config["maximum"])

    assert (
        min_ver_obj <= max_ver_obj
    ), f"Minimum version ({min_ver_obj}) is greater than maximum version in the config file. ({max_ver_obj})"
    assert (
        min_ver_obj <= version_to_check_obj
    ), f"Minimum version ({min_ver_obj}) requested in config is greater than the GaNDLF version ({version_to_check_obj})"
    assert (
        max_ver_obj >= version_to_check_obj
    ), f"Maximum version ({max_ver_obj}) requested in config is less than the GaNDLF version ({version_to_check_obj})"

    return True


def checkPatchDimensions(patch_size: np.ndarray, numlay: int) -> int:
    """
    This function checks the divisibility of a numpy array or integer for architectural integrity

    Args:
        patch_size (np.ndarray): The patch size for checking.
        number (int, optional): The number to check divisibility for. Defaults to 16.

    Returns:
        int: Largest multiple of 2 (less than or equal to numlay) that each element of patch size is divisible by to yield int >= 2
    """
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]

    if all(
        [x >= 2 ** (numlay + 1) and x % 2**numlay == 0 for x in patch_size_to_check]
    ):
        return numlay
    else:
        # base2 = np.floor(np.log2(patch_size_to_check))
        base2 = np.array([getBase2(x) for x in patch_size_to_check])
        remain = patch_size_to_check / 2**base2  # check that at least 1

        layers = np.where(remain == 1, base2 - 1, base2)
        return int(np.min(layers))


def getBase2(num: int) -> int:
    """
    Compute the base 2 logarithm of a number.

    Args:
        num (int): the number

    Returns:
        int: the base 2 logarithm of the number
    """
    # helper for checkPatchDimensions (returns the largest multiple of 2 that num is evenly divisible by)
    base = 0
    while num % 2 == 0:
        num = num / 2
        base = base + 1
    return base


def get_array_from_image_or_tensor(
    input_tensor_or_image: Union[torch.Tensor, sitk.Image]
) -> np.ndarray:
    """
    This function returns the numpy array from a torch.Tensor or sitk.Image.

    Args:
        input_tensor_or_image (Union[torch.Tensor, sitk.Image]): The input tensor or image.

    Returns:
        np.ndarray: The numpy array.
    """
    assert isinstance(
        input_tensor_or_image, (torch.Tensor, sitk.Image, np.ndarray)
    ), "Input must be a torch.Tensor or sitk.Image or np.ndarray, but got " + str(
        type(input_tensor_or_image)
    )
    if isinstance(input_tensor_or_image, torch.Tensor):
        return input_tensor_or_image.detach().cpu().numpy()
    elif isinstance(input_tensor_or_image, sitk.Image):
        return sitk.GetArrayFromImage(input_tensor_or_image)
    elif isinstance(input_tensor_or_image, np.ndarray):
        return input_tensor_or_image


def set_determinism(seed: Optional[int] = 42) -> None:
    """
    This function sets the determinism for the random number generators.

    Args:
        seed (Optional[int], optional): The seed for the random number generators. Defaults to 42.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.use_deterministic_algorithms(True, warn_only=True)

    if torch.cuda.is_available():
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = True


def print_and_format_metrics(
    cohort_level_metrics: dict,
    sample_level_metrics: dict,
    metrics_dict_from_parameters: dict,
    mode: str,
    length_of_dataloader: int,
) -> dict:
    """
    This function prints and formats the metrics.

    Args:
        cohort_level_metrics (dict): The cohort level metrics calculated from the GANDLF.metrics.overall_stats function.
            May be empty dict if not classification/regression.
        sample_level_metrics (dict): The sample level metrics calculated from separate samples from the dataloader(s).
        metrics_dict_from_parameters (dict): The metrics dictionary to populate.
        mode (str): The mode of the metrics (train, val, test).
        length_of_dataloader (int): The length of the dataloader.

    Returns:
        dict: The metrics dictionary populated with the metrics.
    """

    def __update_metric_from_list_to_single_string(input_metrics_dict: dict) -> dict:
        """
        Helper function to update the metric from list to single string.

        Args:
            input_metrics_dict (dict): The input metrics dictionary.

        Returns:
            dict: The output metrics dictionary.
        """
        output_metrics_dict = deepcopy(input_metrics_dict)
        for metric in input_metrics_dict.keys():
            if isinstance(input_metrics_dict[metric], list):
                output_metrics_dict[metric] = ("_").join(
                    str(input_metrics_dict[metric])
                    .replace("[", "")
                    .replace("]", "")
                    .replace(" ", "")
                    .split(",")
                )

        return output_metrics_dict

    output_metrics_dict = deepcopy(cohort_level_metrics)
    for metric in metrics_dict_from_parameters:
        if isinstance(sample_level_metrics[metric], np.ndarray):
            to_print = (sample_level_metrics[metric] / length_of_dataloader).tolist()
        else:
            to_print = sample_level_metrics[metric] / length_of_dataloader
        output_metrics_dict[metric] = to_print
    for metric, metric_val in output_metrics_dict.items():
        print("     Epoch Final   " + mode + " " + metric + " : ", metric_val)
    output_metrics_dict = __update_metric_from_list_to_single_string(
        output_metrics_dict
    )

    return output_metrics_dict


def define_average_type_key(params: dict, metric_name: str) -> str:
    """
    Determine the average type key from the metric config.

    Args:
        params (dict): The parameter dictionary containing training and data information.
        metric_name (str): The name of the metric.

    Returns:
        str: The average type key.
    """
    average_type_key = params["metrics"][metric_name].get("average", "macro")
    return average_type_key


def define_multidim_average_type_key(params: dict, metric_name: str) -> str:
    """
    Determine the multidimensional average type key from the metric config.

    Args:
        params (dict): The parameter dictionary containing training and data information.
        metric_name (str): The name of the metric.

    Returns:
        str: The multidimensional average type key.
    """
    average_type_key = params["metrics"][metric_name].get("multidim_average", "global")
    return average_type_key


def determine_classification_task_type(params: dict) -> str:
    """
    This function determines the classification task type from the parameters.

    Args:
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        str: The classification task type (binary or multiclass).
    """
    task = "binary" if params["model"]["num_classes"] == 2 else "multiclass"
    return task


def get_git_hash() -> str:
    """
    Get the git hash of the current commit.

    Returns:
        str: The git hash of the current commit.
    """
    try:
        git_hash = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=os.getcwd())
            .decode("ascii")
            .strip()
        )
    except (subprocess.CalledProcessError, FileNotFoundError):
        git_hash = "None"
    return git_hash



import sys, os
from typing import List, Optional, Tuple
from pathlib import Path
import numpy as np
import skimage.io

# from skimage.filters.rank import maximum
from skimage.filters import gaussian

# from skimage.morphology.footprints import disk
from skimage.morphology import remove_small_holes
from skimage.color.colorconv import rgb2hsv
import cv2

# from skimage.exposure import rescale_intensity
# from skimage.color import rgb2hed

# import matplotlib.pyplot as plt
import yaml
import openslide

# RGB Masking (pen) constants
RGB_RED_CHANNEL = 0
RGB_GREEN_CHANNEL = 1
RGB_BLUE_CHANNEL = 2
MIN_COLOR_DIFFERENCE = 40

# HSV Masking
HSV_HUE_CHANNEL = 0
HSV_SAT_CHANNEL = 1
HSV_VAL_CHANNEL = 2
MIN_SAT = 20 / 255
MIN_VAL = 30 / 255

# LAB Masking
LAB_L_CHANNEL = 0
LAB_A_CHANNEL = 1
LAB_B_CHANNEL = 2
LAB_L_THRESHOLD = 0.80


def print_sorted_dict(dictionary: dict) -> str:
    """
    Print a dictionary with sorted keys.

    Args:
        dictionary (dict): The input dictionary.

    Returns:
        str: The sorted dictionary.
    """
    sorted_keys = sorted(list(dictionary.keys()))
    output_str = "{"
    for index, key in enumerate(sorted_keys):
        output_str += str(key) + ": " + str(dictionary[key])
        if index < len(sorted_keys) - 1:
            output_str += "; "
    output_str += "}"

    return output_str


def convert_to_tiff(
    filename: str,
    output_dir: str,
    updated_file_name_identifier: Optional[str] = "converted",
) -> str:
    """
    Convert an image to tiff.

    Args:
        filename (str): The input filename.
        output_dir (str): The output directory.
        updated_file_name_identifier (str, optional): The identifier to use for the updated file name. Defaults to "converted".

    Returns:
        str: The path to the converted image.
    """
    base, ext = os.path.splitext(filename)
    # for png or jpg images, write image back to tiff
    if ext in [".png", ".jpg", ".jpeg"]:
        converted_img_path = os.path.join(output_dir, "tiff_converted")
        Path(converted_img_path).mkdir(parents=True, exist_ok=True)
        temp_file = os.path.join(
            converted_img_path,
            os.path.basename(base) + "_" + updated_file_name_identifier + ".tiff",
        )
        temp_img = skimage.io.imread(filename)
        skimage.io.imsave(temp_file, temp_img)
        return temp_file
    else:
        return filename


def pass_method(*args: object) -> str:
    return ""


def get_nonzero_percent(image: np.ndarray) -> float:
    """
    Get the percentage of non-zero pixels in an image.

    Args:
        image (np.ndarray): The input image.

    Returns:
        float: The percentage of non-zero pixels.
    """
    np_img = np.asarray(image)
    non_zero = np.count_nonzero(np_img)
    return non_zero / (np_img.shape[0] * np_img.shape[1])


def get_patch_class_proportions(image: np.ndarray) -> dict:
    """
    Get the class proportions of a patch.

    Args:
        image (np.ndarray): The input image.

    Returns:
        dict: The class proportions
    """
    np_img = np.asarray(image)
    unique, counts = np.unique(image, return_counts=True)
    denom = np_img.shape[0] * np_img.shape[1]
    prop_dict = {val: count / denom for val, count in list(zip(unique, counts))}
    return print_sorted_dict(prop_dict)


def map_values(image: np.ndarray, dictionary: dict) -> np.ndarray:
    """
    Map values in an image to a new set of values.

    Args:
        image (np.ndarray): The input image.
        dictionary (dict): The dictionary to use for mapping.

    Returns:
        np.ndarray: The mapped image.
    """
    template = image.copy()  # Copy image so all values not in dict are unmodified
    for key, value in dictionary.items():
        template[image == key] = value

    return template


## commented because this is GUI and is not used in the code
# def display_overlay(image, mask):
#     overlay = image.copy()
#     overlay[~mask] = (overlay[~mask] // 1.5).astype(np.uint8)
#     plt.imshow(overlay)
#     plt.show()


def hue_range_mask(
    image: np.ndarray, min_hue: float, max_hue: float, sat_min: Optional[float] = 0.05
) -> np.ndarray:
    """
    Mask based on hue range.

    Args:
        image (np.ndarray): RGB numpy image
        min_hue (float): Minimum hue value
        max_hue (float): Maximum hue value
        sat_min (Optional[float], optional): Minimum saturation value. Defaults to 0.05.

    Returns:
        np.ndarray: image mask, True pixels are within the hue range.
    """
    hsv_image = rgb2hsv(image)
    h_channel = gaussian(hsv_image[:, :, HSV_HUE_CHANNEL])
    above_min = h_channel > min_hue
    below_max = h_channel < max_hue

    s_channel = gaussian(hsv_image[:, :, HSV_SAT_CHANNEL])
    above_sat = s_channel > sat_min
    return np.logical_and(np.logical_and(above_min, below_max), above_sat)


def tissue_mask(image: np.ndarray) -> np.ndarray:
    """
    Mask based on low saturation and value (gray-black colors)

    Args:
        image (np.ndarray): RGB numpy image

    Returns:
        np.ndarray: image mask, True pixels are gray-black.
    """
    hue_mask = hue_range_mask(image, 0.8, 0.99)
    final_mask = remove_small_holes(hue_mask)
    return final_mask


### unused function because pen_size_threshold and pen_mask_expansion are not defined
# def basic_pen_mask(image, pen_size_threshold, pen_mask_expansion):
#     green_mask = np.bitwise_and(
#         image[:, :, RGB_GREEN_CHANNEL] > image[:, :, RGB_GREEN_CHANNEL],
#         image[:, :, RGB_GREEN_CHANNEL] - image[:, :, RGB_GREEN_CHANNEL]
#         > MIN_COLOR_DIFFERENCE,
#     )

#     blue_mask = np.bitwise_and(
#         image[:, :, RGB_BLUE_CHANNEL] > image[:, :, RGB_GREEN_CHANNEL],
#         image[:, :, RGB_BLUE_CHANNEL] - image[:, :, RGB_GREEN_CHANNEL]
#         > MIN_COLOR_DIFFERENCE,
#     )

#     masked_pen = np.bitwise_or(green_mask, blue_mask)
#     new_mask_image = remove_small_objects(masked_pen, pen_size_threshold)

#     return maximum(np.where(new_mask_image, 1, 0), disk(pen_mask_expansion)).astype(
#         bool
#     )


### unused function
# def basic_hsv_mask(image):
#     """
#     Mask based on low saturation and value (gray-black colors)
#     :param image: RGB numpy image
#     :return: image mask, True pixels are gray-black.
#     """
#     hsv_image = rgb2hsv(image)
#     return np.bitwise_or(
#         hsv_image[:, :, HSV_SAT_CHANNEL] <= MIN_SAT,
#         hsv_image[:, :, HSV_VAL_CHANNEL] <= MIN_VAL,
#     )


### unused function because pen_size_threshold and pen_mask_expansion are not defined
# def hybrid_mask(image, pen_size_threshold, pen_mask_expansion):
#     return ~np.bitwise_or(basic_hsv_mask(image), basic_pen_mask(image, pen_size_threshold, pen_mask_expansion))


### unused function because pen_size_threshold and pen_mask_expansion are not defined
# def trim_mask(image, mask, pen_size_threshold, pen_mask_expansion, background_value=0, mask_func=hybrid_mask):
#     """
#     Set the values of single-channel image to 0 if outside of whitespace.
#     :param image: RGB numpy image
#     :param mask: Mask to be trimmed
#     :param background_value: Value to set in mask.
#     :param mask_func: Func which takes `image` as a parameter. Returns a binary mask, `True` will be background.
#     :return: `mask` with excess trimmed off
#     """
#     mask_copy = mask.copy()
#     masked_image = mask_func(image)
#     mask_copy[masked_image] = background_value
#     return mask_copy


def patch_size_check(img: np.ndarray, patch_height: int, patch_width: int) -> bool:
    """
    This function checks if the patch size is valid.

    Args:
        img (np.ndarray): Input image.
        patch_height (int): The height of the patch.
        patch_width (int): The width of the patch.

    Returns:
        bool: Whether or not the patch size is valid.
    """
    img = np.asarray(img)

    return_val = False
    if not (img.shape[0] != patch_height or img.shape[1] != patch_width):
        return_val = True

    return return_val


def alpha_rgb_2d_channel_check(img: np.ndarray) -> bool:
    """
    This function checks if an image has a valid alpha channel.

    Args:
        img (np.ndarray): Input image.

    Returns:
        bool: Whether or not the image has alpha channel.
    """
    img = np.asarray(img)
    # If the image has three dimensions AND there is no alpha_channel...
    if len(img.shape) == 3 and img.shape[-1] == 3:
        return True
    # If the image has three dimensions AND there IS an alpha channel...
    elif len(img.shape) == 3 and img.shape[-1] == 4:
        alpha_channel = img[:, :, 3]

        if np.any(alpha_channel != 255):
            return False
        else:
            return True
    # If the image is two dims, return True
    elif len(img.shape) == 2:
        return True
    # Other images (4D, RGBA+____, etc.), return False.
    else:
        return False


# def pen_marking_check(img, intensity_thresh=225, intensity_thresh_saturation =50, intensity_thresh_b = 128):
#    """
#    This function is used to curate patches from the input image. It is used to remove patches that have pen markings.
#    Args:
#        img (np.ndarray): Input Patch Array to check the artifact/background.
#        intensity_thresh (int, optional): Threshold to check whiteness in the patch. Defaults to 225.
#        intensity_thresh_saturation (int, optional): Threshold to check saturation in the patch. Defaults to 50.
#        intensity_thresh_b (int, optional) : Threshold to check blackness in the patch
#        patch_size (int, optional): Tiling Size of the WSI/patch size. Defaults to 256. patch_size=config["patch_size"]
#    Returns:
#        bool: Whether the patch is valid (True) or not (False)
#    """
#    patch_size= (256,256)
#    ihc_hed = rgb2hed(img)
#    patch_hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
#    e = rescale_intensity(ihc_hed[:, :, 1], out_range=(0, 255), in_range=(0, np.percentile(ihc_hed[:, :, 1], 99)))
#    if np.sum(e < 50) / (patch_size[0] * patch_size[1]) > 0.95 or (np.sum(patch_hsv[...,0] < 128) / (patch_size[0] * patch_size[1])) > 0.97:
#        return False
#    #Assume patch is valid
#    return True


def patch_artifact_check(
    img: np.ndarray,
    intensity_thresh: int = 250,
    intensity_thresh_saturation: int = 5,
    intensity_thresh_b: int = 128,
    patch_size: Optional[List[int]] = [256, 256],
) -> bool:
    """
    This function is used to curate patches from the input image. It is used to remove patches that have artifacts.

    Args:
        img (np.ndarray): Input Patch Array to check the artifact/background.
        intensity_thresh (int, optional): Threshold to check whiteness in the patch. Defaults to 250.
        intensity_thresh_saturation (int, optional): Threshold to check saturation in the patch. Defaults to 5.
        intensity_thresh_b (int, optional): Threshold to check blackness in the patch. Defaults to 128.
        patch_size (Optional[List[int]], optional): Tiling Size of the WSI/patch size. Defaults to [256, 256].

    Returns:
        bool: Whether the patch is valid or not.
    """
    # patch_size = config["patch_size"]
    patch_hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    count_white_pixels = np.sum(np.logical_and.reduce(img > intensity_thresh, axis=2))
    percent_pixels = count_white_pixels / (patch_size[0] * patch_size[1])
    count_black_pixels = np.sum(np.logical_and.reduce(img < intensity_thresh_b, axis=2))
    percent_pixel_b = count_black_pixels / (patch_size[0] * patch_size[1])
    percent_pixel_2 = np.sum(patch_hsv[..., 1] < intensity_thresh_saturation) / (
        patch_size[0] * patch_size[1]
    )
    percent_pixel_3 = np.sum(patch_hsv[..., 2] > intensity_thresh) / (
        patch_size[0] * patch_size[1]
    )

    if (
        percent_pixel_2 > 0.99
        or np.mean(patch_hsv[..., 1]) < 5
        or percent_pixel_3 > 0.99
    ):
        if percent_pixel_2 < 0.1:
            return False
    elif (
        (percent_pixel_2 > 0.99 and percent_pixel_3 > 0.99)
        or percent_pixel_b > 0.99
        or percent_pixels > 0.9
    ):
        return False
    # assume that the patch is valid
    return True


def parse_config(config_file: str) -> dict:
    """
    Function that parses the config file.

    Args:
        config_file (str): The path to the config file.

    Returns:
        dict: The parsed config file.
    """
    config = yaml.safe_load(open(config_file, "r"))

    # initialize defaults if not specified
    config["scale"] = config.get("scale", 16)
    config["num_patches"] = config.get("num_patches", -1)
    config["num_workers"] = config.get("num_workers", 1)
    config["save_patches"] = config.get("save_patches", True)
    config["value_map"] = config.get("value_map", None)
    config["read_type"] = config.get("read_type", "random")
    config["overlap_factor"] = config.get("overlap_factor", 0.0)
    config["patch_size"] = config.get("patch_size", [256, 256])

    return config


def is_mask_too_big(mask: np.ndarray) -> bool:
    """
    Function that returns a boolean value indicating whether the mask is too big to make processing slow.

    Args:
        mask (np.ndarray): The valid mask.

    Returns:
        bool: True if the mask is too big, False otherwise.
    """

    if sys.getsizeof(mask) > 16 * (1024**2):
        return True
    else:
        return False


def generate_initial_mask(slide_path: str, scale: int) -> Tuple[np.ndarray, tuple]:
    """
    Function that generates the initial mask for the slide.

    Args:
        slide_path (str): The path to the slide.
        scale (int): The scale to use for the mask.

    Returns:
        Tuple[np.ndarray, tuple]: The valid mask and the real scale.
    """
    # Open slide and get properties
    slide = openslide.open_slide(slide_path)
    slide_dims = slide.dimensions

    # Call thumbnail for effiency, calculate scale relative to whole slide
    slide_thumbnail = np.asarray(
        slide.get_thumbnail((slide_dims[0] // scale, slide_dims[1] // scale))
    )
    real_scale = (
        slide_dims[0] / slide_thumbnail.shape[1],
        slide_dims[1] / slide_thumbnail.shape[0],
    )

    valid_mask = tissue_mask(slide_thumbnail)

    if is_mask_too_big(valid_mask):
        print(
            "Calculated tissue mask is too big; considering increasing the scale for faster processing."
        )

    return valid_mask, real_scale


def get_patch_size_in_microns(
    input_slide_path: str, patch_size_from_config: str, verbose: Optional[bool] = False
) -> List[int]:
    """
    Function that returns the patch size in pixels.

    Args:
        input_slide_path (str): The path to the slide.
        patch_size_from_config (str): The patch size from the config file.
        verbose (Optional[bool], optional): Whether to print verbose output. Defaults to False.

    Returns:
        List[int]: The patch size after getting converted to pixels.
    """
    return_patch_size = [0, 0]
    patch_size = None

    assert isinstance(
        patch_size_from_config, (str, list, tuple)
    ), "Patch size must be a list or string."

    if isinstance(patch_size_from_config, str):
        # first remove all spaces and square brackets
        patch_size_from_config = patch_size_from_config.replace(" ", "")
        patch_size_from_config = patch_size_from_config.replace("[", "")
        patch_size_from_config = patch_size_from_config.replace("]", "")
        # try different split strategies
        patch_size = patch_size_from_config.split(",")
        if len(patch_size) == 1:
            patch_size = patch_size_from_config.split("x")
        if len(patch_size) == 1:
            patch_size = patch_size_from_config.split("X")
        if len(patch_size) == 1:
            patch_size = patch_size_from_config.split("*")
        assert (
            len(patch_size) == 2
        ), "Could not parse patch size from config.yml, use either ',', 'x', 'X', or '*' as separator between x and y dimensions."
    elif isinstance(patch_size_from_config, list) or isinstance(
        patch_size_from_config, tuple
    ):
        patch_size = patch_size_from_config

    magnification_prev = -1
    for i, _ in enumerate(patch_size):
        # for i in range(len(patch_size)):
        magnification = -1
        if str(patch_size[i]).isnumeric():
            return_patch_size[i] = int(patch_size[i])
        elif isinstance(patch_size[i], str):
            if ("m" in patch_size[i]) or ("mu" in patch_size[i]):
                if verbose:
                    print(
                        "Using mpp to calculate patch size for dimension {}".format(i)
                    )
                # only enter if "m" is present in patch size
                input_slide = openslide.open_slide(input_slide_path)
                metadata = input_slide.properties
                if i == 0:
                    for _property in [
                        openslide.PROPERTY_NAME_MPP_X,
                        "tiff.XResolution",
                        "XResolution",
                    ]:
                        if _property in metadata:
                            magnification = float(metadata[_property])
                            magnification_prev = magnification
                            break
                elif i == 1:
                    for _property in [
                        openslide.PROPERTY_NAME_MPP_Y,
                        "tiff.YResolution",
                        "YResolution",
                    ]:
                        if _property in metadata:
                            magnification = float(metadata[_property])
                            break
                    if magnification == -1:
                        # if y-axis data is missing, use x-axis data
                        magnification = magnification_prev
                # get patch size in pixels
                # check for 'mu' first
                size_in_microns = patch_size[i].replace("mu", "")
                size_in_microns = float(size_in_microns.replace("m", ""))
                if verbose:
                    print(
                        "Original patch size in microns for dimension {}",
                        format(size_in_microns),
                    )
                if magnification > 0:
                    return_patch_size[i] = round(size_in_microns / magnification)
                    magnification_prev = magnification
            else:
                return_patch_size[i] = float(patch_size[i])

    if verbose:
        print(
            "Estimated patch size in pixels: [{},{}]".format(
                return_patch_size[0], return_patch_size[1]
            )
        )

    return return_patch_size



#!usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import argparse
import ast

from typing import Optional
from deprecated import deprecated
import click

from GANDLF import version
from GANDLF.cli import main_run, copyrightMessage
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import logger_setup


def _run(
    config: str,
    input_data: str,
    train_flag: bool,
    model_dir: str,
    device: str,
    reset_flag: bool,
    resume_flag: bool,
    output_path: Optional[str],
):
    if model_dir is None and output_path:
        model_dir = output_path

    assert model_dir is not None, "Missing required parameter: model_dir"

    if os.path.isdir(input_data):
        # Is this a fine assumption to make?
        # Medperf models receive the data generated by the data preparator mlcube
        # We can therefore ensure the output of that mlcube contains a data.csv file
        filename = "data.csv"
        input_data = os.path.join(input_data, filename)

    if not train_flag:
        # TODO: print a warning if any of these flags is activated. They are not available for inference mode
        #  Maybe user misconfigured the command.
        # if inference mode, then no need to check for reset/resume
        reset_flag, resume_flag = False, False

    if reset_flag and resume_flag:
        logging.warning(
            "'reset' and 'resume' are mutually exclusive; 'resume' will be used."
        )
        reset_flag = False

    # TODO: check that output_path is not passed in training mode;
    #  maybe user misconfigured the command

    logging.debug(f"{config=}")
    logging.debug(f"{input_data=}")
    logging.debug(f"{train_flag=}")
    logging.debug(f"{model_dir=}")
    logging.debug(f"{device=}")
    logging.debug(f"{reset_flag=}")
    logging.debug(f"{resume_flag=}")
    logging.debug(f"{output_path=}")

    main_run(
        data_csv=input_data,
        config_file=config,
        model_dir=model_dir,
        train_mode=train_flag,
        device=device,
        resume=resume_flag,
        reset=reset_flag,
        output_dir=output_path,
    )
    print("Finished.")


@click.command()
@click.option(
    "--config",
    "-c",
    required=True,
    help="The configuration file (contains all the information related to the training/inference session)",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--input-data",
    "-i",
    required=True,
    type=str,
    help="Data CSV file that is used for training/inference; "
    "can also take comma-separated training-validation pre-split CSVs;"
    "can also take a path to folder with `data.csv`",
)
@click.option(
    "--train/--infer",
    "-t/--infer",
    required=True,  # TODO: what if we make infer as default behavior if no param is passed?
    help="If we run training or inference; for inference, "
    "there needs to be a compatible model saved in '-model-dir'",
)
@click.option(
    "--model-dir",
    "-m",
    type=click.Path(file_okay=False, dir_okay=True),
    help="Training: Output directory to save intermediate files and model weights; "
    "inference: location of previous training session output",
)
@click.option(
    "--device",
    "-d",
    # TODO: Not sure it's worth to restrict this list. What about other devices?
    #  GaNDLF guarantees to work properly with these two options, but
    #  other values may be partially working also.
    #  * GaNDLF code convert `-1` to `cpu` (i.e. it is expected somebody may pass -1)
    #  * `cuda:0` should work also, isn't it? Just would not be treated as `cuda`
    #  * Would `mps` work?
    #  * int values (like `1`) - are they supported? (legacy mode for cuda https://pytorch.org/docs/stable/tensor_attributes.html#torch-device)
    type=click.Choice(["cuda", "cpu"]),
    required=True,  # FIXME: either keep default value, or set required flag
    help="Device to perform requested session on 'cpu' or 'cuda'; "
    "for cuda, ensure CUDA_VISIBLE_DEVICES env var is set",
)
@click.option(
    "--reset",
    "-rt",
    is_flag=True,
    help="Completely resets the previous run by deleting 'model-dir'",
)
@click.option(
    "--resume",
    "-rm",
    is_flag=True,
    help="Resume previous training by only keeping model dict in 'model-dir'",
)
@click.option(
    "--output-path",
    "-o",
    type=click.Path(file_okay=False, dir_okay=True),
    help="Location to save the output of the inference session. Not used for training.",
)
@click.option("--raw-input", hidden=True)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(
    config: str,
    input_data: str,
    train: bool,
    model_dir: str,
    device: str,
    reset: bool,
    resume: bool,
    output_path: str,
    raw_input: str,
    log_file: str,
):
    """Semantic segmentation, regression, and classification for medical images using Deep Learning."""

    logger_setup(log_file)
    _run(
        config=config,
        input_data=input_data,
        train_flag=train,
        model_dir=model_dir,
        device=device,
        reset_flag=reset,
        resume_flag=resume,
        output_path=output_path,
    )


@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf run` cli command "
    + "instead of `gandlf_run`. Note that in new CLI tool some params were renamed or changed its behavior:\n"
    + "  --parameters_file to --config\n"
    + "  --inputdata/--data_path to --input-data\n"
    + "  --train changed its behavior: instead of `--train True/False` pass `--train/--infer`\n"
    + "  --modeldir to --model-dir\n"
    + "  --reset is flag now with default False value if not passed\n"
    + "  --resume is flag now with default False value if not passed\n"
    + "  --outputdir/--output_path to --output-path; in training mode, use --model-dir instead\n"
    + "  --version removed; use `gandlf --version` instead\n"
    + "`gandlf_run` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Semantic segmentation, regression, and classification for medical images using Deep Learning.\n\n"
        + copyrightMessage,
    )
    parser.add_argument(
        "-c",
        "--config",
        "--parameters_file",
        metavar="",
        type=str,
        required=True,
        help="The configuration file (contains all the information related to the training/inference session)",
    )
    parser.add_argument(
        "-i",
        "--inputdata",
        "--data_path",
        metavar="",
        type=str,
        required=True,
        help="Data CSV file that is used for training/inference; can also take comma-separated training-validation pre-split CSVs",
    )
    parser.add_argument(
        "-t",
        "--train",
        metavar="",
        type=ast.literal_eval,
        required=True,
        help="True: training and False: inference; for inference, there needs to be a compatible model saved in '-modeldir'",
    )
    parser.add_argument(
        "-m",
        "--modeldir",
        metavar="",
        type=str,
        help="Training: Output directory to save intermediate files and model weights; inference: location of previous training session output",
    )
    parser.add_argument(
        "-d",
        "--device",
        default="cuda",  # TODO: default value doesn't work as arg is required
        metavar="",
        type=str,
        required=True,
        help="Device to perform requested session on 'cpu' or 'cuda'; for cuda, ensure CUDA_VISIBLE_DEVICES env var is set",
    )
    parser.add_argument(
        "-rt",
        "--reset",
        metavar="",
        default=False,
        type=ast.literal_eval,
        help="Completely resets the previous run by deleting 'modeldir'",
    )
    parser.add_argument(
        "-rm",
        "--resume",
        metavar="",
        default=False,
        type=ast.literal_eval,
        help="Resume previous training by only keeping model dict in 'modeldir'",
    )
    parser.add_argument(
        "-o",
        "--outputdir",
        "--output_path",
        metavar="",
        type=str,
        help="Location to save the output of the inference session. Not used for training.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s v{}".format(version) + "\n\n" + copyrightMessage,
        help="Show program's version number and exit.",
    )

    # This is a dummy argument that exists to trigger MLCube mounting requirements.
    # Do not remove.
    parser.add_argument("-rawinput", "--rawinput", help=argparse.SUPPRESS)

    args = parser.parse_args()

    # config file should always be present
    assert os.path.isfile(args.config), "Configuration file not found!"

    _run(
        config=args.config,
        input_data=args.inputdata,
        train_flag=args.train,
        model_dir=args.modeldir,
        device=args.device,
        reset_flag=args.reset,
        resume_flag=args.resume,
        output_path=args.outputdir,
    )


if __name__ == "__main__":
    old_way()
